@@ -6,7 +6,7 @@
  *  used by the graphics utility of the FreeType test suite.
  *
  *  Written by Antoine Leca.
- *  Copyright (C) 1999-2021 by
+ *  Copyright (C) 1999-2025 by
  *  Antoine Leca, David Turner, Robert Wilhelm, and Werner Lemberg.
  *
  *  Borrowing liberally from the other FreeType drivers.
@@ -19,10 +19,6 @@
  *
  ******************************************************************/
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
 
@@ -40,7 +36,9 @@
 #define  xxDEBUG
 
 #ifdef DEBUG
+# include <stdio.h>
 # include <stdarg.h>
+# include <ctype.h>
 
   static void  LogMessage( const char*  fmt, ... )
   {
@@ -57,7 +55,9 @@
 #endif
 
 /*  Custom messages. */
+#define WM_STATUS  WM_USER+512
 #define WM_RESIZE  WM_USER+517
+#define WM_GR_KEY  WM_USER+519
 
 
   typedef struct  Translator_
@@ -70,18 +70,14 @@
   static
   Translator  key_translators[] =
   {
-    { VK_BACK,      grKeyBackSpace },
-    { VK_TAB,       grKeyTab       },
-    { VK_RETURN,    grKeyReturn    },
-    { VK_ESCAPE,    grKeyEsc       },
+    { VK_PRIOR,     grKeyPageUp    },
+    { VK_NEXT,      grKeyPageDown  },
+    { VK_END,       grKeyEnd       },
     { VK_HOME,      grKeyHome      },
     { VK_LEFT,      grKeyLeft      },
     { VK_UP,        grKeyUp        },
     { VK_RIGHT,     grKeyRight     },
     { VK_DOWN,      grKeyDown      },
-    { VK_PRIOR,     grKeyPageUp    },
-    { VK_NEXT,      grKeyPageDown  },
-    { VK_END,       grKeyEnd       },
     { VK_F1,        grKeyF1        },
     { VK_F2,        grKeyF2        },
     { VK_F3,        grKeyF3        },
@@ -99,6 +95,7 @@
   typedef struct grWin32SurfaceRec_
   {
     grSurface     root;
+    DWORD         host;
     HWND          window;
     HICON         sIcon;
     HICON         bIcon;
@@ -113,8 +110,11 @@
 
 /* destroys the surface*/
 static void
-gr_win32_surface_done( grWin32Surface*  surface )
+gr_win32_surface_done( grSurface*  baseSurface )
 {
+  grWin32Surface*  surface = (grWin32Surface*)baseSurface;
+
+
   /* The graphical window has perhaps already destroyed itself */
   if ( surface->window )
   {
@@ -137,18 +137,20 @@ gr_win32_surface_done( grWin32Surface*  surface )
 
 static void
 gr_win32_surface_refresh_rectangle(
-         grWin32Surface*  surface,
-         int              x,
-         int              y,
-         int              w,
-         int              h )
+         grSurface*  baseSurface,
+         int         x,
+         int         y,
+         int         w,
+         int         h )
 {
-  int        delta;
-  RECT       rect;
-  grBitmap*  bitmap = &surface->root.bitmap;
+  grWin32Surface*  surface = (grWin32Surface*)baseSurface;
+  int              delta;
+  RECT             rect;
+  grBitmap*        bitmap = &surface->root.bitmap;
+
 
   LOG(( "gr_win32_surface_refresh_rectangle: ( %p, %d, %d, %d, %d )\n",
-        (long)surface, x, y, w, h ));
+        surface->root.bitmap.buffer, x, y, w, h ));
 
   /* clip update rectangle */
 
@@ -237,39 +239,25 @@ gr_win32_surface_refresh_rectangle(
 
 
 static void
-gr_win32_surface_set_title( grWin32Surface*  surface,
-                            const char*      title )
+gr_win32_surface_set_title( grSurface*   baseSurface,
+                            const char*  title )
 {
+  grWin32Surface*  surface = (grWin32Surface*)baseSurface;
+
+
   SetWindowText( surface->window, title );
 }
 
 
 static int
-gr_win32_surface_set_icon( grWin32Surface*  surface,
-                           grBitmap*        icon )
+gr_win32_surface_set_icon( grSurface*  baseSurface,
+                           grBitmap*   icon )
 {
-  int       s[] = { GetSystemMetrics( SM_CYSMICON ),
-                    GetSystemMetrics( SM_CYICON ) };
-  WPARAM    wParam;
-  HDC       hDC;
-  VOID*     bts;
-  ICONINFO  ici = { TRUE };
-  HICON     hIcon;
-
-  /* NOTE: The Mingw64 header file `wingdi.h` defines this macro as `sRGB`,
-   * which triggers the `-Wmultichar` warning during compilation, so replace
-   * it with the corresponding numerical value.
-   */
-#ifdef __MINGW64__
-# undef  LCS_sRGB
-# define LCS_sRGB  0x73524742
-#endif
-
-  BITMAPV4HEADER  hdr = { sizeof( BITMAPV4HEADER ),
-                          0, 0, 1, 32, BI_BITFIELDS, 0, 0, 0, 0, 0,
-                          0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000,
-                          LCS_sRGB };
-
+  grWin32Surface*  surface = (grWin32Surface*)baseSurface;
+  int              s[] = { GetSystemMetrics( SM_CYSMICON ),
+                           GetSystemMetrics( SM_CYICON ) };
+  WPARAM           wParam;
+  HICON            hIcon;
 
   if ( !icon )
     return s[1];
@@ -282,19 +270,8 @@ gr_win32_surface_set_icon( grWin32Surface*  surface,
   else
     return 0;
 
-  ici.hbmMask  = CreateBitmap( icon->width, icon->rows, 1, 1, NULL);
-
-  hdr.bV4Width  =  icon->width;
-  hdr.bV4Height = -icon->rows;
-
-  hDC = GetDC( NULL );
-  ici.hbmColor = CreateDIBSection( hDC, (LPBITMAPINFO)&hdr,
-                                   DIB_RGB_COLORS, &bts, NULL, 0 );
-  ReleaseDC( NULL, hDC );
-
-  memcpy( bts, icon->buffer, icon->rows * icon->width * 4 );
-
-  hIcon = CreateIconIndirect( &ici );
+  hIcon = CreateIcon( NULL, icon->width, icon->rows, 1, 32,
+                      NULL, icon->buffer );
 
   PostMessage( surface->window, WM_SETICON, wParam, (LPARAM)hIcon );
 
@@ -305,8 +282,6 @@ gr_win32_surface_set_icon( grWin32Surface*  surface,
     return 0;
   case ICON_BIG:
     surface->bIcon = hIcon;
-    SetClassLongPtr( surface->window, GCLP_HBRBACKGROUND,
-                     (LONG_PTR)CreatePatternBrush( ici.hbmColor ) );
     return s[0];
   default:
     return 0;  /* should not happen */
@@ -320,7 +295,7 @@ gr_win32_surface_set_icon( grWin32Surface*  surface,
  */
 
 
-static grWin32Surface*
+static int
 gr_win32_surface_resize( grWin32Surface*  surface,
                          int              width,
                          int              height )
@@ -343,7 +318,7 @@ gr_win32_surface_resize( grWin32Surface*  surface,
                       width,
                       height,
                       &surface->shadow_bitmap ) )
-    return 0;
+      return 0;
 
 #ifdef SWIZZLE
     if ( grNewBitmap( bitmap->mode,
@@ -361,23 +336,27 @@ gr_win32_surface_resize( grWin32Surface*  surface,
   surface->bmiHeader.biWidth  = width;
   surface->bmiHeader.biHeight = -height;
 
-  return surface;
+  return 1;
 }
 
-static void
-gr_win32_surface_listen_event( grWin32Surface*  surface,
-                               int              event_mask,
-                               grEvent*         grevent )
+static int
+gr_win32_surface_listen_event( grSurface*  baseSurface,
+                               int         event_mask,
+                               grEvent*    grevent )
 {
-  MSG     msg;
+  grWin32Surface*  surface = (grWin32Surface*)baseSurface;
+  MSG              msg;
 
   event_mask=event_mask;  /* unused parameter */
 
-  while ( GetMessage( &msg, NULL, 0, 0 ) > 0 )
+  while ( GetMessage( &msg, (HWND)-1, 0, 0 ) > 0 )
   {
     switch ( msg.message )
     {
     case WM_RESIZE:
+      while( PeekMessage( &msg, (HWND)-1, WM_RESIZE, WM_RESIZE, PM_REMOVE ) )
+        continue;
+
       {
         int  width  = LOWORD(msg.lParam);
         int  height = HIWORD(msg.lParam);
@@ -390,49 +369,83 @@ gr_win32_surface_listen_event( grWin32Surface*  surface,
           grevent->type  = gr_event_resize;
           grevent->x     = width;
           grevent->y     = height;
-          return;
+          return 1;
         }
       }
       break;
 
-    case WM_KEYDOWN:
-    case WM_SYSKEYDOWN:
-      {
-        Translator*  trans = key_translators;
-        Translator*  limit = trans + sizeof( key_translators ) /
-                                     sizeof( key_translators[0] );
-        for ( ; trans < limit; trans++ )
-          if ( msg.wParam == trans->winkey )
-          {
-            grevent->type = gr_event_key;
-            grevent->key  = trans->grkey;
-            LOG(( "KeyPress: VK = 0x%02x\n", msg.wParam ));
-            return;
-          }
-      }
-      break;
-
     case WM_CHAR:
+    case WM_GR_KEY:
       {
         grevent->type = gr_event_key;
         grevent->key  = msg.wParam;
-        LOG(( isprint( msg.wParam ) ? "KeyPress: Char = '%c'\n"
+        LOG(( (msg.wParam <= 256 && isprint( msg.wParam ))
+                                    ? "KeyPress: Char = '%c'\n"
                                     : "KeyPress: Char = <%02x>\n",
               msg.wParam ));
-        return;
+        return 1;
       }
       break;
     }
+  }
 
+  return 0;
+}
+
+
+DWORD WINAPI Window_ThreadProc( LPVOID lpParameter )
+{
+  grWin32Surface*  surface = (grWin32Surface*)lpParameter;
+  DWORD            style   = WS_OVERLAPPEDWINDOW;
+  RECT             WndRect;
+  MSG              msg;
+
+  WndRect.left   = 0;
+  WndRect.top    = 0;
+  WndRect.right  = surface->root.bitmap.width;
+  WndRect.bottom = surface->root.bitmap.rows;
+
+  AdjustWindowRect( &WndRect, style, FALSE );
+
+  surface->window = CreateWindow(
+       /* LPCSTR lpszClassName;    */ "FreeTypeTestGraphicDriver",
+       /* LPCSTR lpszWindowName;   */ "FreeType Test Graphic Driver",
+       /* DWORD dwStyle;           */  style,
+       /* int x;                   */  CW_USEDEFAULT,
+       /* int y;                   */  CW_USEDEFAULT,
+       /* int nWidth;              */  WndRect.right - WndRect.left,
+       /* int nHeight;             */  WndRect.bottom - WndRect.top,
+       /* HWND hwndParent;         */  HWND_DESKTOP,
+       /* HMENU hmenu;             */  NULL,
+       /* HINSTANCE hinst;         */  GetModuleHandle( NULL ),
+       /* void FAR* lpvParam;      */  surface );
+
+  PostThreadMessage( surface->host, WM_STATUS, (WPARAM)surface->window, 0 );
+
+  if ( surface->window == NULL )
+    return -1;
+
+  ShowWindow( surface->window, SW_SHOWNORMAL );
+
+  while ( GetMessage( &msg, NULL, 0, 0 ) > 0 )
+  {
     TranslateMessage( &msg );
     DispatchMessage( &msg );
   }
+
+  LOG(("Window thread done.\n"));
+  return 0;
 }
 
-static grWin32Surface*
-gr_win32_surface_init( grWin32Surface*  surface,
-                       grBitmap*        bitmap )
+
+static int
+gr_win32_surface_init( grSurface*  baseSurface,
+                       grBitmap*   bitmap )
 {
+  grWin32Surface*  surface = (grWin32Surface*)baseSurface;
+  MSG              msg;
+
+
   /* Set default mode */
   if ( bitmap->mode == gr_pixel_mode_none )
   {
@@ -446,64 +459,30 @@ gr_win32_surface_init( grWin32Surface*  surface,
     switch ( bpp )
     {
     case 8:
-      bitmap->mode = gr_pixel_mode_gray;
+      surface->root.bitmap.mode  = gr_pixel_mode_gray;
+      surface->root.bitmap.grays = 256;
       break;
     case 16:
-      bitmap->mode = gr_pixel_mode_rgb565;
+      surface->root.bitmap.mode = gr_pixel_mode_rgb565;
       break;
     case 24:
-      bitmap->mode = gr_pixel_mode_rgb24;
+      surface->root.bitmap.mode = gr_pixel_mode_rgb24;
       break;
     case 32:
     default:
-      bitmap->mode = gr_pixel_mode_rgb32;
+      surface->root.bitmap.mode = gr_pixel_mode_rgb32;
     }
   }
-
-  LOG(( "Win32: init_surface( %p, %p )\n", surface, bitmap ));
-
-  LOG(( "       --   mode   = %d\n", bitmap->mode ));
-  LOG(( "       --   grays  = %d\n", bitmap->grays ));
-  LOG(( "       --   width  = %d\n", bitmap->width ));
-  LOG(( "       --   height = %d\n", bitmap->rows ));
-
-  /* create the bitmap - under Win32, we support all modes as the GDI */
-  /* handles all conversions automatically..                          */
-  if ( grNewBitmap( bitmap->mode,
-                    bitmap->grays,
-                    bitmap->width,
-                    bitmap->rows,
-                    bitmap ) )
-    return 0;
-
-  /* allocate the BGR shadow bitmap */
-  if ( bitmap->mode == gr_pixel_mode_rgb24 )
-  {
-    if ( grNewBitmap( bitmap->mode,
-                      bitmap->grays,
-                      bitmap->width,
-                      bitmap->rows,
-                      &surface->shadow_bitmap ) )
-      return 0;
-
-#ifdef SWIZZLE
-    if ( grNewBitmap( bitmap->mode,
-                      bitmap->grays,
-                      bitmap->width,
-                      bitmap->rows,
-                      &surface->swizzle_bitmap ) )
-      return 0;
-#endif
-  }
   else
-    surface->shadow_bitmap.buffer = bitmap->buffer;
+  {
+    surface->root.bitmap.mode  = bitmap->mode;
+    surface->root.bitmap.grays = bitmap->grays;
+  }
 
   surface->bmiHeader.biSize   = sizeof( BITMAPINFOHEADER );
-  surface->bmiHeader.biWidth  = bitmap->width;
-  surface->bmiHeader.biHeight = -bitmap->rows;
   surface->bmiHeader.biPlanes = 1;
 
-  switch ( bitmap->mode )
+  switch ( surface->root.bitmap.mode )
   {
   case gr_pixel_mode_mono:
     surface->bmiHeader.biBitCount = 1;
@@ -518,17 +497,16 @@ gr_win32_surface_init( grWin32Surface*  surface,
 
   case gr_pixel_mode_gray:
     surface->bmiHeader.biBitCount = 8;
-    surface->bmiHeader.biClrUsed  = bitmap->grays;
+    surface->bmiHeader.biClrUsed  = surface->root.bitmap.grays;
     {
-      int   count = bitmap->grays;
-      int   x;
+      int       x, count = surface->root.bitmap.grays - 1;
       RGBQUAD*  color = surface->bmiColors;
 
-      for ( x = 0; x < count; x++, color++ )
+      for ( x = 0; x <= count; x++, color++ )
       {
         color->rgbRed   =
         color->rgbGreen =
-        color->rgbBlue  = (unsigned char)(x*255/(count-1));
+        color->rgbBlue  = (BYTE)( x * 255 / count );
         color->rgbReserved = 0;
       }
     }
@@ -565,44 +543,37 @@ gr_win32_surface_init( grWin32Surface*  surface,
     return 0;         /* Unknown mode */
   }
 
-  {
-    DWORD  style = WS_OVERLAPPEDWINDOW;
-    RECT   WndRect;
+  /* allocate buffers */
+  if ( !gr_win32_surface_resize( surface, bitmap->width, bitmap->rows ) )
+    goto Fail;
 
-    WndRect.left   = 0;
-    WndRect.top    = 0;
-    WndRect.right  = bitmap->width;
-    WndRect.bottom = bitmap->rows;
+  /* set up the main message queue and spin off the window thread */
+  PeekMessage( &msg, (HWND)-1, WM_USER, WM_USER, PM_NOREMOVE );
+  surface->host = GetCurrentThreadId();
 
-    AdjustWindowRect( &WndRect, style, FALSE );
+  CreateThread( NULL, 0, Window_ThreadProc, (LPVOID)surface, 0, NULL );
 
-    surface->window = CreateWindow(
-            /* LPCSTR lpszClassName;    */ "FreeTypeTestGraphicDriver",
-            /* LPCSTR lpszWindowName;   */ "FreeType Test Graphic Driver",
-            /* DWORD dwStyle;           */  style,
-            /* int x;                   */  CW_USEDEFAULT,
-            /* int y;                   */  CW_USEDEFAULT,
-            /* int nWidth;              */  WndRect.right - WndRect.left,
-            /* int nHeight;             */  WndRect.bottom - WndRect.top,
-            /* HWND hwndParent;         */  HWND_DESKTOP,
-            /* HMENU hmenu;             */  0,
-            /* HINSTANCE hinst;         */  GetModuleHandle( NULL ),
-            /* void FAR* lpvParam;      */  surface );
-  }
+  /* listen if window is created */
+  if ( GetMessage ( &msg, (HWND)-1, WM_STATUS, WM_STATUS ) < 0 ||
+       !msg.wParam )
+    goto Fail;
 
-  if ( surface->window == 0 )
-    return  0;
+  /* wrap up */
+  surface->root.done         = gr_win32_surface_done;
+  surface->root.refresh_rect = gr_win32_surface_refresh_rectangle;
+  surface->root.set_title    = gr_win32_surface_set_title;
+  surface->root.set_icon     = gr_win32_surface_set_icon;
+  surface->root.listen_event = gr_win32_surface_listen_event;
 
-  ShowWindow( surface->window, SW_SHOWNORMAL );
+  LOG(( "Surface initialized: %dx%dx%d\n",
+        surface->root.bitmap.width, surface->root.bitmap.rows,
+        surface->bmiHeader.biBitCount ));
 
-  surface->root.bitmap       = *bitmap;
-  surface->root.done         = (grDoneSurfaceFunc) gr_win32_surface_done;
-  surface->root.refresh_rect = (grRefreshRectFunc) gr_win32_surface_refresh_rectangle;
-  surface->root.set_title    = (grSetTitleFunc)    gr_win32_surface_set_title;
-  surface->root.set_icon     = (grSetIconFunc)     gr_win32_surface_set_icon;
-  surface->root.listen_event = (grListenEventFunc) gr_win32_surface_listen_event;
+  return 1;
 
-  return surface;
+Fail:
+  gr_win32_surface_done( &surface->root );
+  return 0;
 }
 
 
@@ -637,59 +608,56 @@ LRESULT CALLBACK Message_Process( HWND handle, UINT mess,
     {
     case WM_CLOSE:
       /* warn the main thread to quit if it didn't know */
-      PostMessage( handle, WM_CHAR, (WPARAM)grKeyEsc, 0 );
+      PostThreadMessage( surface->host, WM_GR_KEY, (WPARAM)grKeyEsc, 0 );
       break;
 
     case WM_SIZE:
       if ( wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED )
-        PostMessage( handle, WM_RESIZE, wParam, lParam );
+        PostThreadMessage( surface->host, WM_RESIZE, wParam, lParam );
       break;
 
-#ifdef DEBUG
-    case WM_SIZING:
+    case WM_KEYDOWN:
+    case WM_SYSKEYDOWN:
       {
-        PRECT  r = (PRECT)lParam;
-        RECT   WndRect;
-        int    x, y;
+        Translator*  trans = key_translators;
+        Translator*  limit = trans + sizeof( key_translators ) /
+                                     sizeof( key_translators[0] );
 
-        GetClientRect( handle, &WndRect );
-
-        y = wParam >= 6 ? wParam -= 6, 'B' :
-            wParam >= 3 ? wParam -= 3, 'T' : ' ';
-        x = wParam == 2 ? 'R' :
-            wParam == 1 ? 'L' : ' ';
-
-        LOG(( "WM_SIZING %c%c : ( %d %d %d %d )   "
-              "ClientArea : ( %d %d )\n",
-              y, x, r->left, r->top, r->right, r->bottom,
-              WndRect.right, WndRect.bottom ));
-
-        /* XXX: We cannot simply interrupt here and resize the image. */
+        for ( ; trans < limit; trans++ )
+          if ( wParam == trans->winkey )
+          {
+            /* repost to the main thread */
+            PostThreadMessage( surface->host, WM_GR_KEY, trans->grkey, 0 );
+            LOG(( "KeyPress: VK = 0x%02x\n", wParam ));
+            break;
+          }
       }
       break;
-#endif
 
-    case WM_EXITSIZEMOVE:
-      {
-        RECT  WndRect;
-
-        GetClientRect( handle, &WndRect );
-        PostMessage( handle, WM_RESIZE, SIZE_RESTORED,
-                     MAKELPARAM( WndRect.right, WndRect.bottom ) );
-      }
+    case WM_CHAR:
+      /* repost to the main thread */
+      PostThreadMessage( surface->host, mess, wParam, lParam );
       break;
 
     case WM_PAINT:
       {
-        HDC           hDC;
-        PAINTSTRUCT   ps;
+        HDC          hDC;
+        PAINTSTRUCT  ps;
 
-        hDC   = BeginPaint ( handle, &ps );
-        SetDIBitsToDevice( hDC, 0, 0,
-                           surface->root.bitmap.width,
-                           surface->root.bitmap.rows,
-                           0, 0, 0,
-                           surface->root.bitmap.rows,
+        hDC = BeginPaint ( handle, &ps );
+
+        LOG(( "WM_PAINT : ( %d %d %d %d )\n",
+              ps.rcPaint.left,  ps.rcPaint.top,
+              ps.rcPaint.right, ps.rcPaint.bottom ));
+
+        SetDIBitsToDevice( hDC,
+                           ps.rcPaint.left,
+                           ps.rcPaint.top,
+                           ps.rcPaint.right - ps.rcPaint.left,
+                           ps.rcPaint.bottom - ps.rcPaint.top,
+                           ps.rcPaint.left,
+                           surface->root.bitmap.rows - ps.rcPaint.bottom,
+                           0, surface->root.bitmap.rows,
                            surface->shadow_bitmap.buffer,
                            (LPBITMAPINFO)&surface->bmiHeader,
                            DIB_RGB_COLORS );
@@ -711,22 +679,16 @@ LRESULT CALLBACK Message_Process( HWND handle, UINT mess,
       /* WNDPROC lpfnWndProc  */ Message_Process,
       /* int     cbClsExtra   */ 0,
       /* int     cbWndExtra   */ 0,
-      /* HANDLE  hInstance    */ 0,
-      /* HICON   hIcon        */ 0,
-      /* HCURSOR hCursor      */ 0,
-      /* HBRUSH  hbrBackground*/ 0,
+      /* HANDLE  hInstance    */ GetModuleHandle( NULL ),
+      /* HICON   hIcon        */ NULL,
+      /* HCURSOR hCursor      */ LoadCursor( NULL, IDC_ARROW),
+      /* HBRUSH  hbrBackground*/ NULL,
       /* LPCTSTR lpszMenuName */ NULL,
       /* LPCTSTR lpszClassName*/ "FreeTypeTestGraphicDriver"
     };
 
     /* register window class */
-
-    ourClass.hInstance    = GetModuleHandle( NULL );
-    ourClass.hIcon        = LoadIcon(0, IDI_APPLICATION);
-    ourClass.hCursor      = LoadCursor(0, IDC_ARROW);
-    ourClass.hbrBackground= GetStockObject( LTGRAY_BRUSH );
-
-    if ( RegisterClass(&ourClass) == 0 )
+    if ( RegisterClass( &ourClass ) == 0 )
       return -1;
 
     return 0;
@@ -747,7 +709,7 @@ LRESULT CALLBACK Message_Process( HWND handle, UINT mess,
     gr_win32_device_init,
     gr_win32_device_done,
 
-    (grDeviceInitSurfaceFunc) gr_win32_surface_init,
+    gr_win32_surface_init,
 
     0,
     0
