@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright (C) 2007-2021 by                                              */
+/*  Copyright (C) 2007-2025 by                                              */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -17,12 +17,10 @@
 #include "common.h"
 #include "mlgetopt.h"
 
-#include FT_OUTLINE_H
-#include FT_LCD_FILTER_H
-#include FT_DRIVER_H
-
-  /* showing driver name */
-#include FT_MODULE_H
+#include <freetype/ftdriver.h>
+#include <freetype/ftmodapi.h>  /* showing driver name */
+#include <freetype/ftlcdfil.h>
+#include <freetype/ftoutln.h>
 
 #include <stdarg.h>
 #include <stdio.h>
@@ -31,7 +29,7 @@
 
 
   static void
-  usage( char*  execname )
+  usage( const char*  execname )
   {
     fprintf( stderr,
       "\n"
@@ -49,8 +47,7 @@
       "            `.afm' or `.pfm').\n"
       "\n" );
     fprintf( stderr,
-      "  -w W         Set the window width to W pixels (default: 640px).\n"
-      "  -h H         Set the window height to H pixels (default: 480px).\n"
+      "  -d WxH       Set window dimensions (default: 640x480).\n"
       "  -r R         Use resolution R dpi (default: 72dpi).\n"
       "  -s S         Set character size to S points (default: 16pt).\n"
       "  -f TEXTFILE  Change displayed text, using text in TEXTFILE\n"
@@ -180,13 +177,6 @@
     unsigned char  filter_weights[5];
     int            fw_index;
 
-    unsigned int   cff_hinting_engine;
-    unsigned int   type1_hinting_engine;
-    unsigned int   t1cid_hinting_engine;
-    unsigned int   tt_interpreter_versions[3];
-    int            num_tt_interpreter_versions;
-    int            tt_interpreter_version_idx;
-
   } ColumnStateRec, *ColumnState;
 
 
@@ -227,21 +217,6 @@
                      Display      display,
                      FT_Library   library )
   {
-    FT_UInt  cff_hinting_engine;
-    FT_UInt  type1_hinting_engine;
-    FT_UInt  t1cid_hinting_engine;
-
-    unsigned int  tt_interpreter_versions[3]  = { 0, 0, 0 };
-    int           num_tt_interpreter_versions = 0;
-    int           tt_interpreter_version_idx  = 0;
-
-    unsigned int  dflt_tt_interpreter_version;
-    int           i;
-    unsigned int  versions[3] = { TT_INTERPRETER_VERSION_35,
-                                  TT_INTERPRETER_VERSION_38,
-                                  TT_INTERPRETER_VERSION_40 };
-
-
     memset( state, 0, sizeof ( *state ) );
 
     state->library = library;
@@ -254,55 +229,12 @@
     state->char_size    = 16;
     state->display      = display[0];
 
-    /* get the default value as compiled into FreeType */
-    FT_Property_Get( library,
-                     "cff",
-                     "hinting-engine", &cff_hinting_engine );
-    FT_Property_Get( library,
-                     "type1",
-                     "hinting-engine", &type1_hinting_engine );
-    FT_Property_Get( library,
-                     "t1cid",
-                     "hinting-engine", &t1cid_hinting_engine );
-
-    /* collect all available versions, then set again the default */
-    FT_Property_Get( library,
-                     "truetype",
-                     "interpreter-version", &dflt_tt_interpreter_version );
-    for ( i = 0; i < 3; i++ )
-    {
-      error = FT_Property_Set( library,
-                               "truetype",
-                               "interpreter-version", &versions[i] );
-      if ( !error )
-        tt_interpreter_versions[num_tt_interpreter_versions++] = versions[i];
-      if ( versions[i] == dflt_tt_interpreter_version )
-        tt_interpreter_version_idx = i;
-    }
-    FT_Property_Set( library,
-                     "truetype",
-                     "interpreter-version", &dflt_tt_interpreter_version );
-
     state->columns[0].use_cboxes             = 0;
     state->columns[0].use_kerning            = 1;
     state->columns[0].use_deltas             = 1;
     state->columns[0].use_lcd_filter         = 1;
     state->columns[0].lcd_filter             = FT_LCD_FILTER_DEFAULT;
     state->columns[0].hint_mode              = HINT_MODE_BYTECODE;
-    state->columns[0].cff_hinting_engine     = cff_hinting_engine;
-    state->columns[0].type1_hinting_engine   = type1_hinting_engine;
-    state->columns[0].t1cid_hinting_engine   = t1cid_hinting_engine;
-
-    state->columns[0].tt_interpreter_versions[0] =
-      tt_interpreter_versions[0];
-    state->columns[0].tt_interpreter_versions[1] =
-      tt_interpreter_versions[1];
-    state->columns[0].tt_interpreter_versions[2] =
-      tt_interpreter_versions[2];
-    state->columns[0].num_tt_interpreter_versions =
-      num_tt_interpreter_versions;
-    state->columns[0].tt_interpreter_version_idx =
-      tt_interpreter_version_idx;
 
     state->columns[0].use_custom_lcd_filter  = 0;
     state->columns[0].fw_index               = 2;
@@ -383,7 +315,7 @@
   static void
   render_state_set_files( RenderState  state,
                           char**       files,
-                          char*        execname )
+                          const char*  execname )
   {
     FontFace      faces     = NULL;
     unsigned int  num_faces = 0;
@@ -511,20 +443,20 @@
         return -1;
 
       {
-        unsigned int  len = strlen( filepath );
+        unsigned int  len = strlen( filepath ) + 1;
         char*         p;
 
 
-        if ( len + 1 > sizeof ( state->filepath0 ) )
+        if ( len > sizeof ( state->filepath0 ) )
         {
-          state->filepath = (const char*)malloc( len + 1 );
+          state->filepath = (const char*)malloc( len );
           if ( state->filepath == NULL )
           {
             state->filepath = state->filepath0;
             return -1;
           }
         }
-        memcpy( (char*)state->filepath, filepath, len + 1 );
+        memcpy( (char*)state->filepath, filepath, len );
         p = (char*)strrchr( state->filepath, '\\' );
         if ( p == NULL )
           p = (char*)strrchr( state->filepath, '/' );
@@ -567,26 +499,6 @@
     FT_Bool      have_0x0D      = 0;
 
 
-    /* no need to check for errors: the values used here are always valid */
-    FT_Property_Set( state->library,
-                     "cff",
-                     "hinting-engine",
-                     &column->cff_hinting_engine );
-    FT_Property_Set( state->library,
-                     "type1",
-                     "hinting-engine",
-                     &column->type1_hinting_engine );
-    FT_Property_Set( state->library,
-                     "t1cid",
-                     "hinting-engine",
-                     &column->t1cid_hinting_engine );
-
-    FT_Property_Set( state->library,
-                     "truetype",
-                     "interpreter-version",
-                     &column->tt_interpreter_versions
-                       [column->tt_interpreter_version_idx] );
-
     /* changing a property is in most cases a global operation; */
     /* we are on the safe side if we reload the face completely */
     /* (this is something a normal program doesn't need to do)  */
@@ -800,61 +712,40 @@
       extra = "";
       if ( rmode == HINT_MODE_BYTECODE )
       {
-        if ( !strcmp( module_name, "cff" ) )
-        {
-          switch ( column->cff_hinting_engine )
-          {
-          case FT_HINTING_FREETYPE:
-            extra = " (CFF FT)";
-            break;
-          case FT_HINTING_ADOBE:
-            extra = " (CFF Adobe)";
-            break;
-          }
-        }
+        FT_UInt  prop;
 
-        else if ( !strcmp( module_name, "type1" ) )
-        {
-          switch ( column->type1_hinting_engine )
-          {
-          case FT_HINTING_FREETYPE:
-            extra = " (T1 FT)";
-            break;
-          case FT_HINTING_ADOBE:
-            extra = " (T1 Adobe)";
-            break;
-          }
-        }
 
-        else if ( !strcmp( module_name, "t1cid" ) )
+        if ( !FT_Property_Get( state->library, module_name,
+                               "interpreter-version", &prop ) )
         {
-          switch ( column->t1cid_hinting_engine )
-          {
-          case FT_HINTING_FREETYPE:
-            extra = " (CID FT)";
-            break;
-          case FT_HINTING_ADOBE:
-            extra = " (CID Adobe)";
-            break;
-          }
-        }
-
-        else if ( !strcmp( module_name, "truetype" ) )
-        {
-          switch ( column->tt_interpreter_versions[
-                     column->tt_interpreter_version_idx] )
+          switch ( prop )
           {
           case TT_INTERPRETER_VERSION_35:
             extra = " (TT v35)";
             break;
-          case TT_INTERPRETER_VERSION_38:
-            extra = " (TT v38)";
-            break;
           case TT_INTERPRETER_VERSION_40:
             extra = " (TT v40)";
             break;
           }
         }
+
+        else if ( !FT_Property_Get( state->library, module_name,
+                                    "hinting-engine", &prop ) )
+        {
+          switch ( prop )
+          {
+          case FT_HINTING_FREETYPE:
+            extra = !strcmp( module_name, "cff" )   ? " (CFF FT)" :
+                    !strcmp( module_name, "type1" ) ? " (T1 FT)"  :
+                    !strcmp( module_name, "t1cid" ) ? " (CID FT)" : "";
+            break;
+          case FT_HINTING_ADOBE:
+            extra = !strcmp( module_name, "cff" )   ? " (CFF Adobe)" :
+                    !strcmp( module_name, "type1" ) ? " (T1 Adobe)"  :
+                    !strcmp( module_name, "t1cid" ) ? " (CID Adobe)" : "";
+            break;
+          }
+        }
       }
 
       snprintf( temp, sizeof ( temp ), "%s%s",
@@ -1272,42 +1163,48 @@
     case grKEY( 'H' ):
       {
         const char*  module_name = FT_FACE_DRIVER_NAME( state->face );
+        FT_UInt      prop = 0;
 
 
-        if ( column->hint_mode == HINT_MODE_BYTECODE )
+        if ( !FT_Property_Get( state->library, module_name,
+                               "interpreter-version", &prop ) )
         {
-          if ( !strcmp( module_name, "cff" ) )
+          switch ( prop )
           {
-            FTDemo_Event_Cff_Hinting_Engine_Change(
-              state->library,
-              &column->cff_hinting_engine,
-              1 );
+          S1:
+          case TT_INTERPRETER_VERSION_35:
+            prop = TT_INTERPRETER_VERSION_40;
+            if ( !FT_Property_Set( state->library, module_name,
+                                   "interpreter-version", &prop ) )
+              break;
+            /* fall through */
+          case TT_INTERPRETER_VERSION_40:
+            prop = TT_INTERPRETER_VERSION_35;
+            if ( !FT_Property_Set( state->library, module_name,
+                                   "interpreter-version", &prop ) )
+              break;
+            goto S1;
           }
-          else if ( !strcmp( module_name, "type1" ) )
-          {
-            FTDemo_Event_Type1_Hinting_Engine_Change(
-              state->library,
-              &column->type1_hinting_engine,
-              1 );
-          }
-          else if ( !strcmp( module_name, "t1cid" ) )
-          {
-            FTDemo_Event_T1cid_Hinting_Engine_Change(
-              state->library,
-              &column->t1cid_hinting_engine,
-              1 );
-          }
-          else if ( !strcmp( module_name, "truetype" ) )
-          {
-            column->tt_interpreter_version_idx += 1;
-            column->tt_interpreter_version_idx %=
-              column->num_tt_interpreter_versions;
+        }
 
-            FT_Property_Set( state->library,
-                             "truetype",
-                             "interpreter-version",
-                             &column->tt_interpreter_versions[
-                               column->tt_interpreter_version_idx] );
+        else if ( !FT_Property_Get( state->library, module_name,
+                                    "hinting-engine", &prop ) )
+        {
+          switch ( prop )
+          {
+          S2:
+          case FT_HINTING_FREETYPE:
+            prop = FT_HINTING_ADOBE;
+            if ( !FT_Property_Set( state->library, module_name,
+                                   "hinting-engine", &prop ) )
+              break;
+            /* fall through */
+          case FT_HINTING_ADOBE:
+            prop = FT_HINTING_FREETYPE;
+            if ( !FT_Property_Set( state->library, module_name,
+                                   "hinting-engine", &prop ) )
+              break;
+            goto S2;
           }
         }
       }
@@ -1456,8 +1353,8 @@
     const char*     textfile   = NULL;
     char*           text       = (char*)default_text;
 
-    char*  execname;
-    int    option;
+    const char*  execname;
+    int          option;
 
 
     execname  = ft_basename( argv[0] );
@@ -1467,23 +1364,23 @@
 
     while ( 1 )
     {
-      option = getopt( argc, argv, "f:h:r:s:vw:" );
+      option = getopt( argc, argv, "d:f:r:s:v" );
 
       if ( option == -1 )
         break;
 
       switch ( option )
       {
+      case 'd':
+        if ( sscanf( optarg, "%dx%d", &width, &height ) != 2 ||
+             width < 1 || height < 1 )
+          usage( execname );
+       break;
+
       case 'f':
         textfile = optarg;
         break;
 
-      case 'h':
-        height = atoi( optarg );
-        if ( height < 1 )
-          usage( execname );
-        break;
-
       case 'r':
         resolution = atoi( optarg );
         break;
@@ -1507,12 +1404,6 @@
         }
         /* break; */
 
-      case 'w':
-        width = atoi( optarg );
-        if ( width < 1 )
-          usage( execname );
-        break;
-
       default:
         usage( execname );
         break;
@@ -1629,9 +1520,10 @@
                          column_width, column_height );
 
       write_global_info( state );
-      grRefreshSurface( adisplay->surface );
 
+      grRefreshSurface( adisplay->surface );
       grListenSurface( adisplay->surface, 0, &event );
+
       if ( event.type == gr_event_resize )
       {
         width  = event.x;
